from xiaoaitts.lib.invoke import ubus


def pause(ticket):
    message = {
        'action': 'pause',
        'media': 'app_ios'
    }
    return ubus(ticket, message=message, method='player_play_operation', path='mediaplayer')



from xiaoaitts.player.get_context import get_context
from xiaoaitts.player.get_playlist import get_playlist
from xiaoaitts.player.get_playlist_songs import get_playlist_songs
from xiaoaitts.player.get_song_info import get_song_info
from xiaoaitts.player.get_status import get_status
from xiaoaitts.player.get_volume import get_volume
from xiaoaitts.player.next import next
from xiaoaitts.player.pause import pause
from xiaoaitts.player.play import play
from xiaoaitts.player.play_album_playlist import play_album_playlist
from xiaoaitts.player.play_url import play_url
from xiaoaitts.player.prev import prev
from xiaoaitts.player.set_play_loop import set_play_loop
from xiaoaitts.player.set_volume import set_volume
from xiaoaitts.player.toggle_play_state import toggle_play_state

VOLUME_STEP = 5


def volume_up(ticket):
    volume = get_volume(ticket)
    return set_volume(ticket, volume + VOLUME_STEP)


def volume_down(ticket):
    volume = get_volume(ticket)
    return set_volume(ticket, volume - VOLUME_STEP)


def get_my_playlist(ticket, list_id=None):
    playlist = get_playlist(ticket)
    if not list_id:
        return playlist
    song_list = None
    for item in playlist:
        if item['id'] == int(list_id):
            song_list = item
            break
    show_count = song_list['songCount'] if song_list else 0
    return get_playlist_songs(ticket, list_id=list_id, count=show_count)



import json

from xiaoaitts.lib.invoke import ubus


def nlp(ticket):
    data = ubus(ticket, message={}, method='nlp_result_get', path='mibrain')
    info = json.loads(data['info'])
    return json.loads(info['result'][0]['nlp'])



from xiaoaitts.const import get_model_spec
from xiaoaitts.lib.invoke import miot


def tts(ticket, text):
    model_spec = get_model_spec(ticket.hardware)
    params = {
        'did': ticket.miot_did,
        'siid': model_spec['siid'],
        'aiid': model_spec['aiid'],
        'in': [text]
    }
    return miot(ticket, cmd='action', params=params)



APP_DEVICE_ID = '3C861A5820190429'
SDK_VER = '3.9'
APP_UA = 'APP/com.xiaomi.mihome APPV/6.0.103 iosPassportSDK/3.9.0 iOS/14.4 miHSTS'
MINA_UA = 'MiHome/6.0.103 (com.xiaomi.mihome; build:6.0.103.1; iOS 14.4.0) Alamofire/6.0.103 MICO/iOSApp/appStore/6.0.103'
MIIO_UA = 'iOS-14.4-6.0.103-iPhone12,3--D7744744F7AF32F0544445285880DD63E47D9BE9-8816080-84A3F44E137B71AE-iPhone'

MODEL_SPECS = {
    # 已验证：lx01=小爱迷你音箱，lx5a=小爱音箱Play，lx04=小爱触屏音箱，x08c=红米小爱音箱
    # 未测试：l04m，l04n，l05c，l06a，lx05，lx06，s12，x08a
    'lx04,l04m,x08a': {'execute_aiid': 4},
    'l04n,l05c': {'aiid': 3, 'execute_aiid': 4},
    'x08c': {'siid': 3, 'execute_siid': 3, 'volume_siid': 4},

    # TODO: '123': {'aiid': 3, 'execute_aiid': 4}, # 未知型号，直接执行文本参数只有一个，应该不能工作，需要调整代码，但可能市面上不存在这个型号，先忽略
    # TODO：'v1,v3': # 不支持 MIoT TTS，需要引入 MiNA Service 支持，我手上没有相关设备，如果有人需要可以提出
}


def get_model_spec(model):
    default = {
        'siid': 5,
        'aiid': 1,
        'execute_siid': 5,
        'execute_aiid': 5,
        'volume_siid': 2,
        'volume_piid': 1
    }
    if model:
        model = model.lower().split('.')[-1]
        for k in MODEL_SPECS:
            if model in k:
                default.update(MODEL_SPECS[k])
    return default


class API:
    MIIO = 'https://api.io.mi.com/app'
    USBS = 'https://api2.mina.mi.com/remote/ubus'
    SERVICE_AUTH = 'https://account.xiaomi.com/pass/serviceLoginAuth2'
    SERVICE_LOGIN = 'https://account.xiaomi.com/pass/serviceLogin'
    PLAYLIST = 'https://api2.mina.mi.com/music/playlist/v2/lists'
    PLAYLIST_SONGS = 'https://api2.mina.mi.com/music/playlist/v2/songs'
    SONG_INFO = 'https://api2.mina.mi.com/music/song_info'
    DEVICE_LIST = 'https://api2.mina.mi.com/admin/v2/device_list'



from xiaoaitts.const import API
from xiaoaitts.lib.invoke import invoke


def get_song_info(ticket, song_id):
    data = {
        'songId': song_id
    }
    return invoke(url=API.SONG_INFO, data=data, cookies=ticket.cookies)



import json

import requests

from xiaoaitts.const import MINA_UA, APP_UA, MIIO_UA


class HttpError(Exception):

    def __init__(self, response) -> None:
        self.url = response.url
        self.status = response.status_code
        self.response = response.text
        self.message = '\n'.join([
            'Request Error',
            'url: %s' % self.url,
            'status: %d' % self.status,
            'response: %s' % self.response
        ])

    def __str__(self) -> str:
        return self.message


def request(url, data, method='GET', response_type='json', headers=None, cookies=None, **kwargs):
    method = method.upper()
    content_type = 'application/x-www-form-urlencoded' if 'POST' == method else 'application/json'

    _headers = {
        'Content-Type': content_type,
        'Connection': 'keep-alive',
        'User-Agent': MINA_UA if 'mina.mi.com' in url else MIIO_UA if 'io.mi.com' in url else APP_UA,
        'Accept': '*/*',
    }
    _headers.update(headers or {})

    if method == 'GET':
        resp = requests.get(url, params=data, headers=_headers, cookies=cookies, **kwargs)
    else:
        content_type = _headers['Content-Type']
        if 'application/json' in content_type:
            resp = requests.post(url, json=data, headers=_headers, cookies=cookies, **kwargs)
        else:
            resp = requests.post(url, data=data, headers=_headers, cookies=cookies, **kwargs)

    if resp and resp.status_code == 200:
        if response_type == 'raw':
            return resp
        elif response_type == 'json':
            return json.loads(resp.text.replace('&&&START&&&', ''))
        else:
            return resp.text

    raise HttpError(resp)



import base64
from hashlib import sha1, md5

from xiaoaitts import XiaoAiError, ErrorCode
from xiaoaitts.const import APP_DEVICE_ID, SDK_VER, API
from xiaoaitts.lib.request import request


class Ticket:

    def __init__(self, **entries):
        self.__dict__.update(entries)

    def __repr__(self):
        return str(self.__dict__)


def login(sid, user, password):
    if not user or not password:
        raise XiaoAiError(ErrorCode.INVALID_INPUT)
    sign = get_login_sign(sid)
    auth_info = service_auth(sid, sign, user, password)
    if auth_info['code'] != 0:
        raise XiaoAiError(auth_info['code'], auth_info['desc'])
    service_token = login_mi_ai(auth_info)
    session = {
        'service_token': service_token,
        'user_id': auth_info['userId'],
        'device_id': APP_DEVICE_ID,
        'ssecurity': auth_info['ssecurity'],
    }
    session['cookies'] = get_cookie(**session)
    return session


def get_cookie(user_id=None, service_token=None, device_id=None, serial_number=None, **kwargs):
    cookies = {
        'userId': str(user_id or ''),
        'serviceToken': service_token or '',
    }
    if device_id and service_token:
        cookies.update({
            'deviceId': device_id,
            'sn': serial_number,
        })
    return cookies


def get_login_sign(sid):
    data = {
        'sid': sid,
        '_json': True
    }
    info = request(url=API.SERVICE_LOGIN, data=data)
    return {'_sign': info['_sign'], 'qs': info['qs'], 'callback': info['callback']}


def service_auth(sid, sign, user, password):
    data = {
        'user': user,
        'hash': md5(password.encode()).hexdigest().upper(),
        **sign,
        'sid': sid,
        '_json': True
    }
    cookies = {
        'deviceId': APP_DEVICE_ID,
        'sdkVersion': SDK_VER,
    }
    return request(url=API.SERVICE_AUTH, method='POST', data=data, cookies=cookies)


def login_mi_ai(auth_info):
    def gen_client_sign():
        s = 'nonce={nonce}&{ssecurity}'.format_map(auth_info)
        return base64.b64encode(sha1(s.encode()).digest()).decode()

    data = {
        'clientSign': gen_client_sign()
    }
    try:
        resp = request(url=auth_info['location'], data=data, response_type='raw')
        for cookie in resp.cookies:
            if cookie.name == 'serviceToken':
                return cookie.value
    except Exception as e:
        raise XiaoAiError(e)
    return ''


def switch_session_device(session, device):
    session.update({
        'device_id': device['deviceID'],
        'serial_number': device['serialNumber'],
        'miot_did': device['miotDID'],
        'hardware': device['hardware'],
    })
    session['cookies'] = get_cookie(**session)
    return session



from xiaoaitts.lib.invoke import ubus


def set_volume(ticket, volume):
    volume = min(max(volume, 0), 100)
    message = {
        'volume': volume,
    }
    return ubus(ticket, message=message, method='player_set_volume', path='mediaplayer')



from xiaoaitts.lib.invoke import ubus


def toggle_play_state(ticket):
    message = {
        'action': 'toggle',
    }
    return ubus(ticket, message=message, method='player_play_operation', path='mediaplayer')



from xiaoaitts import player
from xiaoaitts.get_device import get_device
from xiaoaitts.lib.error import XiaoAiError, ErrorCode
from xiaoaitts.lib.invoke import ubus
from xiaoaitts.login import login, switch_session_device, Ticket
from xiaoaitts.tts import tts


class XiaoAi:

    def __init__(self, user, password):
        self.current_device = None
        self.mina_session = login('micoapi', user, password)
        devices = get_device(self.mina_session['cookies'])
        if len(devices) == 0:
            return
        self.current_device = devices[0]
        self.mina_session = switch_session_device(self.mina_session, self.current_device)

        self.miio_session = login('xiaomiio', user, password)
        self.miio_session = switch_session_device(self.miio_session, self.current_device)

    def test(self):
        return ubus(Ticket(**self.mina_session),
                    message={'vendor_name': 'XiaoMi_M88'},
                    method='tts_vendor_switch',
                    path='mibrain')

    def get_device(self, name=None):
        devices = get_device(self.mina_session['cookies'])
        if not name:
            return devices
        target = []
        for device in devices:
            if name in device['name']:
                target.append(device)
        return target or None

    def use_device(self, device_id):
        session = self.mina_session
        devices = get_device(session['cookies'])
        device = None
        for _device in devices:
            if _device['deviceID'] == device_id:
                device = _device
                break
        if not device:
            raise XiaoAiError(ErrorCode.NO_DEVICE, device_id)
        self.current_device = device
        self.mina_session = switch_session_device(session, device)
        self.miio_session = switch_session_device(self.miio_session, device)

    def _call(self, method, **kwargs):
        if method.__name__ == 'tts':
            ticket = Ticket(**self.miio_session)
        else:
            ticket = Ticket(**self.mina_session)
        return method(ticket, **kwargs)

    def say(self, text):
        return self._call(tts, text=text)

    def set_volume(self, volume):
        if isinstance(volume, int):
            raise XiaoAiError(ErrorCode.INVALID_INPUT)
        return self._call(player.set_volume, volume=volume)

    def get_volume(self):
        return self._call(player.get_volume)

    def volume_up(self):
        return self._call(player.volume_up)

    def volume_down(self):
        return self._call(player.volume_down)

    def play(self):
        return self._call(player.play)

    def pause(self):
        return self._call(player.pause)

    def prev(self):
        return self._call(player.prev)

    def next(self):
        return self._call(player.next)

    def set_play_loop(self, type=1):
        return self._call(player.set_play_loop, type=type)

    def get_status(self):
        return self._call(player.get_status)

    def get_song_info(self, song_id):
        return self._call(player.get_song_info, song_id=song_id)

    def get_my_playlist(self, list_id=None):
        return self._call(player.get_my_playlist, list_id=list_id)



from xiaoaitts.const import API
from xiaoaitts.lib.invoke import invoke


def get_device(cookies):
    data = {
        'master': 0
    }
    devices = invoke(url=API.DEVICE_LIST, data=data, cookies=cookies)
    live_devices = []
    for device in devices:
        if device['presence'] == 'online':
            live_devices.append(device)
    return live_devices



from xiaoaitts.lib.invoke import ubus


def play_album_playlist(ticket, album_playlist_id, type=1, start_offset=1):
    message = {
        'type': type,
        'id': album_playlist_id,
        'startOffset': start_offset,
        'media': 'common'
    }
    return ubus(ticket, message=message, method='player_play_album_playlist', path='mediaplayer')
